﻿using System;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;


namespace BeeDevelopment.BbcBasicEditor.Wpf {
	/// <summary>
	/// Interaction logic for MainInterface.xaml
	/// </summary>
	public partial class MainInterface : Window {

		/// <summary>
		/// Gets the application's product name.
		/// </summary>
		private string ApplicationName {
			get {
				return ((AssemblyProductAttribute)(Assembly.GetExecutingAssembly().GetCustomAttributes(typeof(AssemblyProductAttribute), false)[0])).Product;
			}
		}
		
		public MainInterface() {
			InitializeComponent();
			this.OpenDialog = new System.Windows.Forms.OpenFileDialog();
			this.SaveDialog = new System.Windows.Forms.SaveFileDialog();
			this.Editor.Focus();
			this.UpdateTitle();
		}

		#region Dialogs

		private System.Windows.Forms.OpenFileDialog OpenDialog;
		private System.Windows.Forms.SaveFileDialog SaveDialog;

		#endregion

		#region File Readers/Writers

		public void WriteAllText(string filename, string text) {
			this.CurrentFile.SetDataFromString(text ?? "");
			this.CurrentFile.Name = Path.GetFileNameWithoutExtension(filename).ToUpperInvariant();
			this.CurrentFile.Save(filename);
		}

		public string ReadAllText(string filename) {
			try {
				this.CurrentFile = BbcBasicFile.FromFile(filename);
				return this.CurrentFile.GetDataAsString();
			} catch {
				try {
					return Tokeniser.Detokenise(File.ReadAllBytes(filename));
				} catch {
					return File.ReadAllText(filename);
				}
			}
		}

		#endregion
		
		private BbcBasicFile CurrentFile { get; set; }

		#region Text Editing File Stuff

		private string currentFilename;
		/// <summary>
		/// Gets or sets the name of the file currently being edited.
		/// </summary>
		public string CurrentFilename {
			get { return this.currentFilename; }
			set {
				this.currentFilename = value;
				this.UpdateTitle();
			}
		}

		private bool currentFileIsDirty;
		/// <summary>
		/// Gets or sets whether the current file is dirty.
		/// </summary>
		public bool CurrentFileIsDirty {
			get { return this.currentFileIsDirty; }
			set {
				this.currentFileIsDirty = value;
				this.UpdateTitle();
			}
		}

		/// <summary>
		/// Updates the title of the form.
		/// </summary>
		private void UpdateTitle() {
			string Title = "";
			if (this.CurrentFilename != null) {
				Title = Path.GetFileName(this.CurrentFilename);
			} else {
				Title = "Untitled";
			}
			if (this.CurrentFileIsDirty) Title += "*";
			Title += " - " + ApplicationName;
			this.Title = Title;
		}
		
		/// <summary>
		/// Saves a file.
		/// </summary>
		/// <returns>True if the file was saved, false otherwise.</returns>
		public bool Save() {
			if (this.CurrentFilename == null) {
				return this.SaveAs();
			} else {
				try {
					this.WriteAllText(this.CurrentFilename, this.Editor.Text);
					this.CurrentFileIsDirty = false;
					return true;
				} catch (Exception ex) {
					if (MessageBox.Show(this, "Error saving file. " + ex.Message + Environment.NewLine + "Would you like to try saving in a different location?", ApplicationName, MessageBoxButton.YesNo, MessageBoxImage.Error) == MessageBoxResult.Yes) {
						return this.SaveAs();
					} else {
						return false;
					}
				}
			}
		}

		/// <summary>
		/// Saves a file as...
		/// </summary>
		/// <returns>True if the file was saved, false otherwise.</returns>
		public bool SaveAs() {
			if (this.SaveDialog.ShowDialog() == System.Windows.Forms.DialogResult.Cancel) {
				return false;
			}
			try {
				this.WriteAllText(this.SaveDialog.FileName, this.Editor.Text);
				this.CurrentFilename = this.SaveDialog.FileName;
				this.CurrentFileIsDirty = false;
				return true;
			} catch (Exception ex) {
				if (MessageBox.Show(this, "Error saving file. " + ex.Message + Environment.NewLine + "Would you like to try saving in a different location?", ApplicationName, MessageBoxButton.YesNo, MessageBoxImage.Error) == MessageBoxResult.Yes) {
					return this.SaveAs();
				} else {
					return false;
				}
			}
		}

		/// <summary>
		/// Opens a file.
		/// </summary>
		/// <param name="filename">The name of the file to open. Passing null prompts to open using a common file dialog.</param>
		public void Open(string filename) {
			if (this.CurrentFileIsDirty) {
				switch (MessageBox.Show(this, "Would you like to save the current file before opening another?", ApplicationName, MessageBoxButton.YesNoCancel)) {
					case MessageBoxResult.Cancel:
						return;
					case MessageBoxResult.Yes:
						if (!this.Save()) return;
						break;
					case MessageBoxResult.No:
						break;
				}
			}
			try {
				if (filename == null && this.OpenDialog.ShowDialog() == System.Windows.Forms.DialogResult.Cancel) return;
				filename = filename ?? this.OpenDialog.FileName;
				this.Editor.Text = this.ReadAllText(filename);
				this.CurrentFilename = filename;
				this.CurrentFileIsDirty = false;
				//this.UpdateCursorStatus();
			} catch (Exception ex) {
				MessageBox.Show(this, "Error opening file: " + ex.Message, ApplicationName, MessageBoxButton.OK, MessageBoxImage.Error);
			}
		}
		
		#endregion

		private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e) {
			if (this.CurrentFileIsDirty) {
				switch (MessageBox.Show(this, "Would you like to save the current file before exiting?", ApplicationName, MessageBoxButton.YesNoCancel)) {
					case MessageBoxResult.Cancel:
						e.Cancel = true;
						return;
					case MessageBoxResult.Yes:
						if (!this.Save()) {
							e.Cancel = true;
							return;
						}
						break;
					case MessageBoxResult.No:
						break;
				}
			}
		}

		private void Window_Closed(object sender, EventArgs e) {
			if (this.OpenDialog != null) { this.OpenDialog.Dispose(); this.OpenDialog = null; }
			if (this.SaveDialog != null) { this.SaveDialog.Dispose(); this.SaveDialog = null; }
		}

		private void Editor_TextChanged(object sender, TextChangedEventArgs e) {
			this.CurrentFileIsDirty = true;
		}

		#region Commands

		public static RoutedCommand ToggleWordWrap = new RoutedCommand();

		#endregion

		#region Command Handlers

		// A command that can always be executed can use this.
		private void CmdCanAlwaysExecute(object sender, CanExecuteRoutedEventArgs e) { e.CanExecute = true; }

		private void NewCmdExecuted(object sender, ExecutedRoutedEventArgs e) {
			if (this.CurrentFileIsDirty) {
				switch (MessageBox.Show("Would you like to save the current file before creating a new one?", ApplicationName, MessageBoxButton.YesNoCancel, MessageBoxImage.None)) {
					case MessageBoxResult.Cancel:
						return;
					case MessageBoxResult.Yes:
						if (!this.Save()) return;
						break;
					case MessageBoxResult.No:
						break;
				}
			}
			this.Editor.Text = "";
			this.CurrentFilename = null;
			this.CurrentFileIsDirty = false;
		}

		private void OpenCmdExecuted(object sender, ExecutedRoutedEventArgs e) {
			this.Open(null);
		}

		private void SaveCmdExecuted(object sender, ExecutedRoutedEventArgs e) {
			this.Save();
		}

		private void SaveAsCmdExecuted(object sender, ExecutedRoutedEventArgs e) {
			this.SaveAs();
		}

		private void ToggleWordWrapCmdExecuted(object sender, ExecutedRoutedEventArgs e) {
			this.Editor.TextWrapping = this.Editor.TextWrapping == TextWrapping.NoWrap ? TextWrapping.Wrap : TextWrapping.NoWrap;			
		}

		#endregion

		private void FindCmdExecuted(object sender, ExecutedRoutedEventArgs e) {
			new FindAndReplace().ShowDialog();
		}

	}
}
